!!.. this module contains VTI one-way extrapolator with optimized coefficients
!!--------------------------------------------------------------------
!!.. Author: Yunyue Li, yunyue@sep.stanford.edu
module ani_extrpl_mod
  use  base_fftwv3_mod
  implicit none

contains
  !!---------------------------------------------------------------------------------------------------------
  !!.. Phase-shift extrapolator
  !!---------------------------------------------------------------------------------------------------------
  integer function ssr_phs(adj, wfld_slice, vmod_slice, kx2, ky2, vref, alpha0, beta0, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI) result(stat)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real   , dimension(:)   :: kx2, ky2

    real                    :: vref, alpha0, beta0
    real                    :: ws
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign
    integer*8               :: plan_2dxF
    integer*8               :: plan_2dxI

    stat = 1

    if (adj) then
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
       call phase_shift(adj, wfld_slice, kx2, ky2, vref, alpha0, beta0, ws, dz, nx, ny, opsign)
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
    else
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny) 
       call phase_shift(adj, wfld_slice, kx2, ky2, vref, alpha0, beta0, ws, dz, nx, ny, opsign)
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
    end if
    stat = 0
  end function

  !!-----------------------------------------------------------------------------------------------------------
  !!.. Frequency domain implicit finite-difference extrapolator
  !!.. Dip upto 15, 45, 65, 79, 80, 87, 89, 90
  !!-----------------------------------------------------------------------------------------------------------
  integer function ssr_ifd(adj, wfld_slice, vmod_slice, alpha, beta, kx2, ky2, vref, eref, dref, ws, alpha0, beta0, dx, dy, dz, &
				   trick, bndp, nx, ny, dip, opsign, bullet, errflt, rwrk1d, cwrk1d, plan_2dxF, plan_2dxI) result(stat)
    logical                          :: adj
    complex, dimension(:,:)          :: wfld_slice
    real   , dimension(:,:)          :: vmod_slice
    real   , dimension(:,:)          :: alpha 
    real   , dimension(:,:)          :: beta
    real   , dimension(:)            :: kx2, ky2
    real                             :: ws, alpha0, beta0
    real                             :: vref, eref, dref
    real                             :: dx, dy, dz, trick
    complex                          :: bndp
    integer                          :: nx, ny, dip
    integer, intent(in)              :: opsign
    logical                          :: bullet, errflt
    real   , dimension(nx+ny)        :: rwrk1d
    real   , dimension(nx+ny)        :: awrk1d
    real   , dimension(nx+ny)        :: bwrk1d
    complex, dimension(nx*5+ny*5)    :: cwrk1d

    integer*8                        :: plan_2dxF
    integer*8                        :: plan_2dxI

    integer                          :: ix, iy, istep, nstep
    integer                          :: lrx1, lry1
    integer                          :: lcx1, lcx2, lcx3, lcx4, lcx5
    integer                          :: lcy1, lcy2, lcy3, lcy4, lcy5

    !!.. asign work space
    lcx1 = 1            !!.. wfdx
    lcx2 = lcx1 + nx    !!.. ahx
    lcx3 = lcx2 + nx    !!.. bhx
    lcx4 = lcx3 + nx    !!.. chx
    lcx5 = lcx4 + nx    !!.. rhx
    lcy1 = lcx5 + nx    !!.. wfdy
    lcy2 = lcy1 + ny    !!.. ahy
    lcy3 = lcy2 + ny    !!.. bhy
    lcy4 = lcy3 + ny    !!.. chy
    lcy5 = lcy4 + ny    !!.. rhy
    
    lrx1 = 1            !!.. velx
    lry1 = lrx1 + nx    !!.. vely

    if (adj) then

!      apply error filter if requested
!      -------------------------------
       if (errflt) then
          call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
          call ifd_errfltr(adj, wfld_slice, kx2, ky2, alpha, beta, trick, ws, vref, eref, dref, &
                           dx, dy, dz, nx, ny, nstep, opsign)
          call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
       end if

!      solver for diffraction term in y first
!     ---------------------------------------
       if (ny > 1) then
          do ix=1, nx
             cwrk1d(lcy1:lcy2-1) = wfld_slice(ix,:)
             rwrk1d(lry1:nx+ny ) = vmod_slice(ix,:)
             awrk1d(lry1:nx+ny ) = alpha     (ix,:)
             bwrk1d(lry1:nx+ny ) = beta      (ix,:)
             call ifd_sysform(adj, cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), rwrk1d(lry1), ws, dy, dz, &
                                 awrk1d(lry1), bwrk1d(lry1), trick, bndp, ny, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcy1), cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), cwrk1d(lcy5), ny)
             wfld_slice(ix,:) = cwrk1d(lcy1:lcy2-1)
          end do
       end if

!      solve for diffraction term in x
!      ------------------------------- 
       if (nx > 1) then
          do iy=1, ny
             cwrk1d(lcx1:lcx2-1) = wfld_slice(:,iy)
             rwrk1d(lrx1:lry1-1) = vmod_slice(:,iy)
             awrk1d(lrx1:lry1-1) = alpha     (:,iy)
             bwrk1d(lrx1:lry1-1) = beta      (:,iy)
             call ifd_sysform(adj, cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), rwrk1d(lrx1), ws, dx, dz, &
                                 awrk1d(lrx1), bwrk1d(lrx1), trick, bndp, nx, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcx1), cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), cwrk1d(lcx5), nx)
             wfld_slice(:,iy) = cwrk1d(lcx1:lcx2-1)
          end do
       end if

!      solve for the thin lens term
!      ----------------------------
       call ifd_sftterm(adj, wfld_slice, vmod_slice, ws, dz, nx, ny, opsign)

    else

!      solve for the thin lens term
!      ----------------------------
       call ifd_sftterm(adj, wfld_slice, vmod_slice, ws, dz, nx, ny, opsign)
!      solve for the diffraction term in x first
!      -----------------------------------------
       if (nx > 1) then
          do iy=1, ny
             cwrk1d(lcx1:lcx2-1) = wfld_slice(:,iy)
             rwrk1d(lrx1:lry1-1) = vmod_slice(:,iy)
             awrk1d(lrx1:lry1-1) = alpha     (:,iy)
             bwrk1d(lrx1:lry1-1) = beta      (:,iy)
             call ifd_sysform(adj, cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), rwrk1d(lrx1), ws, dx, dz, &
                                 awrk1d(lrx1), bwrk1d(lrx1), trick, bndp, nx, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcx1), cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), cwrk1d(lcx5), nx)
             wfld_slice(:,iy) = cwrk1d(lcx1:lcx2-1)
          end do
       end if

!      solve for the diffraction term in y
!      ----------------------------------- 
       if (ny > 1) then
          do ix=1, nx
             cwrk1d(lcy1:lcy2-1) = wfld_slice(ix,:)
             rwrk1d(lry1:nx+ny ) = vmod_slice(ix,:)
             awrk1d(lry1:nx+ny ) = alpha     (ix,:)
             bwrk1d(lry1:nx+ny ) = beta      (ix,:)
             call ifd_sysform(adj, cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), rwrk1d(lry1), ws, dy, dz, &
                             awrk1d(lry1), bwrk1d(lry1), trick, bndp, ny, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcy1), cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), cwrk1d(lcy5), ny)
             wfld_slice(ix,:) = cwrk1d(lcy1:lcy2-1)
          end do
       end if

!      apply error filter, if requested
!      --------------------------------
       if (errflt) then
          call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
          call ifd_errfltr(adj, wfld_slice, kx2, ky2, alpha, beta, trick, ws, vref, eref, dref,&
                           dx, dy, dz, nx, ny, nstep, opsign)
          call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
       end if
    end if
    stat = 0
  end function

  !!-------------------------------------------------------------------------------------------
  !!.. solving for diffraction term
  !!------------------------------------------------------------------------------------------- 
  subroutine ifd_dfrterm(adj, wfdh, ah, bh, ch, rh, nh)
    logical                :: adj
    complex, dimension(nh) :: wfdh
    complex, dimension(nh) :: ah, bh, ch, rh
    integer                :: nh

    integer                :: ih

    rh = 0.

    if (adj) then
       call ctridiag(ah,bh,ch,wfdh,rh,nh)
       do ih=2, nh-1
          wfdh(ih) = conjg(ah(ih))*rh(ih-1) &
                   + conjg(bh(ih))*rh(ih  ) &
                   + conjg(ch(ih))*rh(ih+1)
       end do
       wfdh(1)  = conjg(bh(1))*rh(1) &
                + conjg(ch(1))*rh(2)
       wfdh(nh) = conjg(ah(nh))*rh(nh-1) &
                + conjg(bh(nh))*rh(nh  )

    else
       ! form the right hand side of the equation
       do ih=2, nh-1
          rh(ih) = conjg(ah(ih))*wfdh(ih-1) &
                 + conjg(bh(ih))*wfdh(ih  ) &
                 + conjg(ch(ih))*wfdh(ih+1)
       end do
       ! boundary points
       rh(1)  = conjg(bh(1))*wfdh(1) &
              + conjg(ch(1))*wfdh(2)
       rh(nh) = conjg(ah(nh))*wfdh(nh-1) &
              + conjg(bh(nh))*wfdh(nh  )
       ! tri-diagonal solver
       call ctridiag(ah,bh,ch,rh,wfdh,nh)
    end if
    return
  end subroutine


  !!--------------------------------------------------------------------------
  !!.. solving for shifting term
  !!--------------------------------------------------------------------------
  subroutine ifd_sftterm(adj, wfld_slice, vmod_slice, ws, dz, nx, ny, opsign)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real                    :: ws, dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign

    integer                 :: ix, iy, psign
    real                    :: wov, shift

    if (adj) then
       psign = -opsign
    else
       psign =  opsign
    end if   

    do iy=1, ny
    do ix=1, nx
       wov = ws / vmod_slice(ix,iy) 
       shift = psign * wov * dz 
       wfld_slice(ix,iy) = wfld_slice(ix,iy) * cmplx(cos(shift), sin(shift))
    end do 
    end do

    return
  end subroutine

  !!---------------------------------------------------------------------------
  !!.. finite-difference error compensation filter
  !!.. Li, Z., Compensating finite-difference errors in 3-D migration and 
  !!.. modeling, 1991, VOL. 56, Geophysics. 
  !!---------------------------------------------------------------------------
!  subroutine ifd_errfltr(adj, wfld_slice, kx2, ky2, trick, ws, vavr, eavr, davr, &
!                         dx, dy, dz, nx, ny, nstep, opsign)
!    logical                 :: adj
!    complex, dimension(:,:) :: wfld_slice
!    real   , dimension(:)   :: kx2, ky2
!    real  				    :: alph, beta
!    real                    :: trick
!    real                    :: ws, vavr, eavr, davr
!    real                    :: dx, dy, dz 
!    integer                 :: nx, ny, nstep, opsign
!
!    integer                 :: ix, iy, istep
!    real                    :: sx, sy, wov, laplx, laply
!    real                    :: kz2, kz, phase, kx, ky
!    complex                 :: spltx, splty
!	real					:: aax, bbx, aay, bby
!
!	alph = 0.47824206
!	beta = 0.37636953
!
!    wov = ws / vavr 
!
!    if (adj) then
!       do iy=1, ny
!       do ix=1, nx
!          sx = kx2(ix)/wov/wov
!          sy = ky2(iy)/wov/wov
!          kx  = sqrt(kx2(ix))  !! this sign is not important, since cosine function is an even function
!          ky  = sqrt(ky2(iy))
!          laplx = (-2.+2.*cos(kx*dx))/dx/dx
!          laply = (-2.+2.*cos(ky*dy))/dy/dy
!!         imaginary coefficients
!          aax = opsign*dz*0.5*alph/wov
!          aay = opsign*dz*0.5*alph/wov
!!         real coefficients 
!          bbx = trick*dx*dx + beta/wov/wov
!          bby = trick*dy*dy + beta/wov/wov
!!         calculate the dispersion relation for FD scheme
!          spltx = (1 + cmplx(bbx,aax)*laplx) / (1 + cmplx(bbx,-aax)*laplx)
!          splty = (1 + cmplx(bby,aay)*laplx) / (1 + cmplx(bby,-aay)*laplx)
!!		  Original kz2		  
!          kz2 = 1. - (sx + sy)
!          if (kz2>=0.) then
!             wfld_slice(ix,iy) = wfld_slice(ix,iy)*conjg(spltx)*conjg(splty)
!          else
!             wfld_slice(ix,iy) = 0.
!          end if
!       end do
!       end do
!
!    else
!
!
!       do iy=1, ny
!       do ix=1, nx
!          sx = kx2(ix)/wov/wov 
!          sy = ky2(iy)/wov/wov
!          kx  = sqrt(kx2(ix))  !! this sign is not important, since cosine function is an even function
!          ky  = sqrt(ky2(iy))
!          laplx = (-2.+2.*cos(kx*dx))/dx/dx
!          laply = (-2.+2.*cos(ky*dy))/dy/dy
!!         imaginary coefficients
!          aax = opsign*dz*0.5*alph/wov
!          aay = opsign*dz*0.5*alph/wov
!!         real coefficients 
!          bbx = trick*dx*dx + beta/wov/wov
!          bby = trick*dy*dy + beta/wov/wov
!!         calculate the dispersion relation for FD scheme
!          spltx = (1 + cmplx(bbx,aax)*laplx) / (1 + cmplx(bbx,-aax)*laplx)
!          splty = (1 + cmplx(bby,aay)*laplx) / (1 + cmplx(bby,-aay)*laplx)
!! 		  original kz2		  
!          kz2 = 1. - (sx + sy) 
!          if (kz2>=0. ) then
!             wfld_slice(ix,iy) = wfld_slice(ix,iy)*spltx*splty
!          else
!             wfld_slice(ix,iy) = 0. 
!          end if
!       end do
!       end do
!    end if
!    return
!  end subroutine

  subroutine ifd_errfltr(adj, wfld_slice, kx2, ky2, alph, beta, trick, ws, vref, eref, dref,&
                         dx, dy, dz, nx, ny, nstep, opsign)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:)   :: kx2, ky2
    real   , dimension(:,:)	:: alph, beta
    real                    :: trick
    real                    :: ws, vref, eref, dref
    real                    :: dx, dy, dz 
    integer                 :: nx, ny, nstep, opsign

    integer                 :: ix, iy, istep
    real                    :: sx, sy, wov, laplx, laply
    real                    :: kz2, kz, phase, kx, ky, fa, fb
    complex                 :: spltx, splty, aax, bbx, aay, bby

    wov = ws / vref 
	fa = 0.5
	fb = 0.25

!   alph = 0.5181678      
!	beta = 0.5960820 

    if (adj) then
       do iy=1, ny
       do ix=1, nx
          sx = -kx2(ix)/wov/wov
          sy = -ky2(iy)/wov/wov
          kx  = sqrt(kx2(ix))  !! this sign is not important, since cosine function is an even function
          ky  = sqrt(ky2(iy))
          laplx = (2.-2.*cos(kx*dx))/dx/dx
          laply = (2.-2.*cos(ky*dy))/dy/dy
          spltx = cmplx(1.,0.)
          splty = cmplx(1.,0.)
!         imaginary coefficients
          aax = opsign*cmplx(0.,1.)*dz*0.5*fa/wov
          aay = opsign*cmplx(0.,1.)*dz*0.5*fa/wov
!         real coefficients 
          bbx = trick*dx*dx + fb/wov/wov
          bby = trick*dy*dy + fb/wov/wov
!         calculate the dispersion relation for FD scheme
          spltx = spltx*(1.-(bbx+aax)*laplx)/(1.-(bbx-aax)*laplx)
          splty = splty*(1.-(bby+aay)*laply)/(1.-(bby-aay)*laply)
          kz2 = (1. + (sx+sy)*(1+2*eref)) / (1 + 2*(sx+sy)*(eref-dref))
          if (kz2>=0.) then
             kz = sqrt(kz2)
             phase = opsign*wov*(kz-1)*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*conjg(cmplx(cos(phase),sin(phase)))
             wfld_slice(ix,iy) = wfld_slice(ix,iy)/conjg(spltx)/conjg(splty)
          else
!			 if (kz2<-1) kz2 = -1	  
             kz = sqrt(-kz2)
             phase = -wov*kz*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*0.0!exp(phase)
          end if
       end do
       end do

    else
       do iy=1, ny
       do ix=1, nx
          sx = -kx2(ix)/wov/wov
          sy = -ky2(iy)/wov/wov
          kx  = sqrt(kx2(ix))  !! this sign is not important, since cosine function is an even function
          ky  = sqrt(ky2(iy))
          laplx = (2.-2.*cos(kx*dx))/dx/dx
          laply = (2.-2.*cos(ky*dy))/dy/dy
          spltx = cmplx(1.,0.)
          splty = cmplx(1.,0.)
!         imaginary coefficients
          aax = opsign*cmplx(0.,1.)*dz*0.5*fa/wov
          aay = opsign*cmplx(0.,1.)*dz*0.5*fa/wov
!         real coefficients 
          bbx = trick*dx*dx + fb/wov/wov
          bby = trick*dy*dy + fb/wov/wov
!         calculate the dispersion relation for FD scheme
          spltx = spltx*(1.-(bbx+aax)*laplx)/(1.-(bbx-aax)*laplx)
          splty = splty*(1.-(bby+aay)*laply)/(1.-(bby-aay)*laply)
          kz2 = (1. + (sx+sy)*(1+2*eref)) / (1 + 2*(sx+sy)*(eref-dref))
          if (kz2>=0.) then
             kz = sqrt(kz2)
             phase = opsign*wov*(kz-1)*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*cmplx(cos(phase),sin(phase))
             wfld_slice(ix,iy) = wfld_slice(ix,iy)/spltx/splty
          else
!			 if (kz2<-1) kz2 = -1	  
             kz = sqrt(-kz2)
             phase = -wov*kz*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*0.0!exp(phase)
          end if
       end do
       end do
    end if
    return
  end subroutine

  !!---------------------------------------------------------------------------------------
  !!.. Forming the tridiagonal matrix for implicit finite-difference extrapolation
  !!---------------------------------------------------------------------------------------
  subroutine ifd_sysform(adj, ah, bh, ch, velh, ws, dh, dz, alph, beta, trick, bndp, nh, opsign, bullet)
    logical                :: adj
    complex, dimension(nh) :: ah, bh, ch
    real   , dimension(nh) :: velh
    real   , dimension(nh) :: alph
    real   , dimension(nh) :: beta
    real                   :: ws, dh, dz
    real                   :: trick
    complex                :: bndp
    integer                :: nh, opsign
    logical                :: bullet

    integer                :: ih
    real                   :: velm, velo, velp
    complex                :: aam, aao, aap, bbm, bbo, bbp

    ah = 0; bh = 0; ch = 0

    if (bullet) then

       do ih=1, nh
          velo = velh(ih)
          velm = velh(max(1   , ih-1))
          velp = velh(min(ih+1, nh  ))
!         imaginary coefficients 
          aam  = opsign*cmplx(0.,1.)*dz*0.5*alph(ih) /ws*sqrt(velo)*sqrt(velm)/dh/dh
          aao  = opsign*cmplx(0.,1.)*dz*0.5*alph(ih) /ws*sqrt(velo)*sqrt(velo)/dh/dh
          aap  = opsign*cmplx(0.,1.)*dz*0.5*alph(ih) /ws*sqrt(velo)*sqrt(velp)/dh/dh
!         real coefficients 
          bbm  = trick + beta(ih) /ws/ws*velo*velm/dh/dh
          bbo  = trick + beta(ih) /ws/ws*velo*velo/dh/dh
          bbp  = trick + beta(ih) /ws/ws*velo*velp/dh/dh 
!         sub-diagoal
          ah(ih) = bbm - aam
!         super-diagonal
          ch(ih) = bbp - aap
!         diagonal
          bh(ih) = 1 - 2.*(bbo-aao)
!         boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1 - (1+bndp)*(bbo-aao)
       end do

    else

       do ih=1, nh
          velo = velh(ih)
!         imaginary coefficients
          aao  = opsign*cmplx(0.,1.)*dz*0.5*alph(ih) /ws*sqrt(velo)*sqrt(velo)/dh/dh
!         real coefficients
          bbo  = trick + beta(ih) /ws/ws*velo*velo/dh/dh
!         sub-diagoal
          ah(ih) = bbo - aao
!         super-diagonal
          ch(ih) = bbo - aao
!         diagonal
          bh(ih) = 1 - 2.*(bbo-aao)
!         boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1 - (1+bndp)*(bbo-aao)
       end do

    end if  

    if (adj) then
!      transpose and conjugate the matrix
       do ih=1, nh-1
          ah(1)    = ah(ih+1)
          ah(ih+1) = ch(ih)
          ch(ih)   = ah(1)
       end do
       ah = conjg(ah)
       bh = conjg(bh)
       ch = conjg(ch)
    end if

!   not used elements
    ah(1)  = 0.
    ch(nh) = 0.

    return 
  end subroutine

 
  !!-----------------------------------------------------------------------
  !!.. Phase-shift kernel
  !!-----------------------------------------------------------------------
  subroutine phase_shift(adj, wfld_slice, kx2, ky2, vref, alpha, beta, ws, dz, nx, ny, opsign)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:)   :: kx2
    real   , dimension(:)   :: ky2
    real                    :: vref, ws, alpha, beta 
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign

    integer                 :: ix, iy, psign
    real                    :: kz, kz2, up, down
    real                    :: phase, wov

    if (adj) then
       psign = -opsign
    else 
       psign =  opsign
    end if


    wov = ws / vref

    do iy=1, ny
    do ix=1, nx
	   up   =    alpha * (kx2(ix) + ky2(ix)) / wov / wov
	   down = 1 - beta * (kx2(ix) + ky2(ix)) / wov / wov
       kz   = 1 - up / down 
       phase = psign*wov*kz*dz 
       wfld_slice(ix,iy) = wfld_slice(ix,iy)*cmplx(cos(phase),sin(phase))
    end do
    end do
    return
  end subroutine


  !!-------------------------------------------------------------------------
  !!.. Split-step correction
  !!-------------------------------------------------------------------------
  subroutine ssf_correct(adj, wfld_slice, vmod_slice, vref, ws, dz, nx, ny, opsign)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real                    :: vref
    real                    :: ws
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign
 
    integer                 :: ix, iy, psign
    real                    :: split, wov

    if (adj) then
       psign = -opsign
    else
       psign =  opsign
    end if
    wov = ws / vref
    
    do iy=1, ny
    do ix=1, nx
       split = psign*wov*dz*(vref/vmod_slice(ix,iy)-1.)
       wfld_slice(ix,iy) = wfld_slice(ix,iy)*cmplx(cos(split),sin(split))
    end do 
    end do
    return
  end subroutine


  !!---------------------------------------------------------------------------
  !!.. complex tridiagonal solver
  !!---------------------------------------------------------------------------
  !!.. a(n), sub-diagonal elemnets, only elements from 2:n are used
  !!.. b(n), diagonal elemetns
  !!.. c(n), super-diagonal elements, only elements from 1:n-1 are used
  !!
  !!   /  b(1), c(1)                                     \ / u(  1) \     / r(  1) \ 
  !!   |  a(2), b(2), c(2)                               | | u(  2) |     | r(  2) |
  !!   |        a(3), b(3), c(3)                         | | u(  3) |  =  | r(  3) |
  !!   |              ... ...                            | |  ...   |     |  ...   |
  !!   |                          a(n-1), b(n-1), c(n-1) | | u(n-1) |     | r(n-1) |
  !!   \                                  a(  n), b(  n) / \ u(  n) /     \ r(  n) /
! ************************
! NOTE: r will be modified
! ************************
  subroutine ctridiag(a,b,c,r,u,n)
    integer :: n
    complex :: a(n), b(n), c(n), r(n), u(n)
    integer :: j
    complex :: bet

    bet  = b(1)
    u(1) = r(1)/bet
    do j=2, n
       r(j-1) = c(j-1)/bet
       bet    = b(j)-a(j)*r(j-1)
       u(j) = (r(j)-a(j)*u(j-1))/bet
    end do
    do j=n-1, 1, -1
       u(j) = u(j) - r(j)*u(j+1)
    end do
  end subroutine

! ************************
! nothing will be modified
! ************************
  subroutine ctridiag2(a,b,c,r,u,n)
    integer            :: n
    complex            :: a(n), b(n), c(n), r(n), u(n)
    integer            :: j
    integer, parameter :: NMAX = 5000
    complex            :: bet, gam(NMAX)
    !if (cabs(b(1)) == 0.) stop "ctirdag: rewrite equations"
    bet  = b(1)
    u(1) = r(1)/bet
    do j=2, n
       gam(j) = c(j-1)/bet
       bet    = b(j)-a(j)*gam(j)
       !if (cabs(bet) == 0.) stop "ctridag failed"
       u(j) = (r(j)-a(j)*u(j-1))/bet
    end do 
    do j=n-1,1,-1
       u(j) = u(j) - gam(j+1)*u(j+1)
    end do
    return
  end subroutine

! ************************
! b, c, u will be modified
! ************************
  subroutine ctridiag3(a,b,c,u,n)
    integer :: n
    complex :: a(n), b(n), c(n), u(n)
    integer :: i
    c(1) = c(1)/b(1)
    u(1) = u(1)/b(1)
    do i=2,n
       b(i) = b(i) - a(i)*c(i-1)
       if (i<n) then
          c(i) = c(i)/b(i)
       end if
       u(i) = (u(i)-a(i)*u(i-1))/b(i)
    end do
    do i=n-1,1,-1
       u(i) = u(i) - c(i)*u(i+1)
    end do
    return
  end subroutine

  integer function rsign(x)
    real :: x
    real :: threshold
    threshold = abs(epsilon(x))
    if (abs(x) <= threshold) then
       rsign = 0
    else if (x > +threshold) then
       rsign = +1
    else if (x < -threshold) then
       rsign = -1
    end if
  end function

end module
