module hesinv_dipflt_mod
  use base_struct_mod
  use base_hlxtyp_mod
  use bwi_poldiv_mod
  use bwi_hlxcon_mod
  implicit none

  type(axis)                                , private :: ax, ay, az
  type(helix), dimension(:)    , pointer    , private :: dfcx
  type(helix), dimension(:)    , pointer    , private :: dfcy
  real       , dimension(:)    , pointer    , private :: dfcx0
  real       , dimension(:)    , pointer    , private :: dfcy0
  integer    , dimension(:,:,:), pointer    , private :: selx
  integer    , dimension(:,:,:), pointer    , private :: sely
  
  real       , dimension(:,:)  , allocatable, private :: tmpx
  real       , dimension(:,:)  , allocatable, private :: tmpy
  real       , dimension(:)    , allocatable, private :: tmpi
  logical                                   , private :: smooth
contains


  integer function hesinv_dipflt_init(dfcx_in, dfcy_in, dfcx0_in, dfcy0_in, &
                                      selx_in, sely_in, ax_in, ay_in, az_in, smooth_in) result(stat)
    type(helix), dimension(:)    , target :: dfcx_in
    type(helix), dimension(:)    , target :: dfcy_in
    real       , dimension(:)    , target :: dfcx0_in
    real       , dimension(:)    , target :: dfcy0_in
    integer    , dimension(:,:,:), target :: selx_in
    integer    , dimension(:,:,:), target :: sely_in
    type(axis)                            :: ax_in, ay_in, az_in
    logical                               :: smooth_in

    stat = 1
    dfcx   => dfcx_in
    dfcy   => dfcy_in
    dfcx0  => dfcx0_in
    dfcy0  => dfcy0_in
    selx   => selx_in
    sely   => sely_in
    ax     =  ax_in 
    ay     =  ay_in
    az     =  az_in
    smooth = smooth_in
    allocate(tmpx(ax%n,az%n))
    allocate(tmpy(ay%n,az%n))
    allocate(tmpi(ax%n*ay%n*az%n))
    stat = 0
  end function

  integer function hesinv_dipflt_clos() result(stat)
    stat = 1
    deallocate(tmpx, tmpy, tmpi) 
    stat = 0
  end function

  integer function hesinv_dipflt_oper(adj, add, modl, data) result(stat)
    logical, intent(in) :: adj, add
    real, dimension(:)  :: modl, data

    stat = 1
    if (.not.add) then
       if (adj) then
          modl = 0.
       else
          data = 0.
       end if 
    end if

    if (adj) then
       tmpi = data
       if (smooth) then
          call dipflt(tmpi)
       else 
          call dipkil(tmpi)
       end if
       modl = modl + tmpi
    else
       tmpi = modl
       if (smooth) then 
          call dipflt(tmpi)
       else
          call dipkil(tmpi)
       end if
       data = data + tmpi
    end if 

    stat = 0
  end function


! symmetric operator, forward and adjoint are the same 
  subroutine dipflt(imag) 
    real, dimension(ax%n,ay%n,az%n) :: imag
    integer                         :: ix, iy, stat

!   Ax^{-1}Ay^{-1}Ay^{-T}Ax^{-T}
    if (ax%n>1) then
       stat = bwi_poldiv_init(dfcx, dfcx0, ax%n, az%n)
       do iy=1, ay%n
          stat = bwi_poldiv_oper(.true. ,.false.,tmpx,imag(:,iy,:),selx(:,iy,:))
          imag(:,iy,:) = tmpx
       end do
       stat = bwi_poldiv_clos()
    end if

    if (ay%n>1) then
       stat = bwi_poldiv_init(dfcy, dfcy0, ay%n, az%n)
       do ix=1, ax%n
          stat = bwi_poldiv_oper(.true. ,.false.,tmpy,imag(ix,:,:),sely(ix,:,:))
          stat = bwi_poldiv_oper(.false.,.false.,tmpy,imag(ix,:,:),sely(ix,:,:))
       end do
       stat = bwi_poldiv_clos()
    end if

    if (ax%n>1) then
       stat = bwi_poldiv_init(dfcx, dfcx0, ax%n, az%n)
       do iy=1, ay%n
          tmpx = imag(:,iy,:)
          stat = bwi_poldiv_oper(.false.,.false.,tmpx,imag(:,iy,:),selx(:,iy,:))
       end do
       stat = bwi_poldiv_clos()
    end if
    return
  end subroutine

  subroutine dipkil(imag)
    real, dimension(ax%n,ay%n,az%n)     :: imag
    integer                             :: ix, iy, stat

    real, dimension(:,:,:), allocatable :: temp

    allocate(temp(ax%n,ay%n,az%n))

!   apply Ax^T Ax
    stat = bwi_hlxcon_init(dfcx, dfcx0, ax%n, az%n)
    do iy=1, ay%n
       stat = bwi_hlxcon_oper(.false.,.false.,imag(:,iy,:),tmpx,selx(:,iy,:))
       stat = bwi_hlxcon_oper(.true. ,.false.,temp(:,iy,:),tmpx,selx(:,iy,:)) 
    end do
    stat = bwi_hlxcon_clos()

!   apply Ay^T Ay
    stat = bwi_hlxcon_init(dfcy, dfcy0, ay%n, az%n)
    do ix=1, ax%n
       stat = bwi_hlxcon_oper(.false.,.false.,imag(ix,:,:),tmpy,sely(ix,:,:))
       stat = bwi_hlxcon_oper(.true. ,.false.,imag(ix,:,:),tmpy,selx(ix,:,:))
    end do
    stat = bwi_hlxcon_clos()

    imag = imag + temp

    deallocate(temp)
    return
  end subroutine

end module
